#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.core module of the BEAT platform.             #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

"""
======
docker
======

Execution utilities
"""

import os
import shutil
import requests
import simplejson
import zmq
import docker

import logging
logger = logging.getLogger(__name__)

from .. import stats
from .. import message_handler
from .. import utils

from .remote import RemoteExecutor


class DockerExecutor(RemoteExecutor):
    """DockerExecutor runs the code given an execution block information, externally


    Parameters:

      host (:py:class:`.dock.Host`): A configured docker host that will
        execute the user process. If the host does not have access to the
        required environment, an exception will be raised.

      prefix (str): Establishes the prefix of your installation.

      data (dict, str): The piece of data representing the block to be executed.
        It must validate against the schema defined for execution blocks. If a
        string is passed, it is supposed to be a fully qualified absolute path to
        a JSON file containing the block execution information.

      cache (:py:class:`str`, Optional): If your cache is not located under
        ``<prefix>/cache``, then specify a full path here. It will be used
        instead.

      dataformat_cache (:py:class:`dict`, Optional): A dictionary mapping
        dataformat names to loaded dataformats. This parameter is optional and,
        if passed, may greatly speed-up database loading times as dataformats
        that are already loaded may be re-used. If you use this parameter, you
        must guarantee that the cache is refreshed as appropriate in case the
        underlying dataformats change.

      database_cache (:py:class:`dict`, Optional): A dictionary mapping
        database names to loaded databases. This parameter is optional and, if
        passed, may greatly speed-up database loading times as databases that
        are already loaded may be re-used. If you use this parameter, you must
        guarantee that the cache is refreshed as appropriate in case the
        underlying databases change.

      algorithm_cache (:py:class:`dict`, Optional): A dictionary mapping
        algorithm names to loaded algorithms. This parameter is optional and,
        if passed, may greatly speed-up database loading times as algorithms
        that are already loaded may be re-used. If you use this parameter, you
        must guarantee that the cache is refreshed as appropriate in case the
        underlying algorithms change.

      library_cache (:py:class:`dict`, Optional): A dictionary mapping library
        names to loaded libraries. This parameter is optional and, if passed,
        may greatly speed-up library loading times as libraries that are
        already loaded may be re-used. If you use this parameter, you must
        guarantee that the cache is refreshed as appropriate in case the
        underlying libraries change.


    Attributes:

      cache (str): The path to the cache currently being used

      errors (list): A list containing errors found while loading this execution
        block.

      data (dict): The original data for this executor, as loaded by our JSON
        decoder.

      algorithm (.algorithm.Algorithm): An object representing the algorithm to
        be run.

      databases (dict): A dictionary in which keys are strings with database
        names and values are :py:class:`.database.Database`, representing the
        databases required for running this block. The dictionary may be empty
        in case all inputs are taken from the file cache.

      views (dict): A dictionary in which the keys are tuples pointing to the
        ``(<database-name>, <protocol>, <set>)`` and the value is a setup view
        for that particular combination of details. The dictionary may be empty
        in case all inputs are taken from the file cache.

      input_list (beat.backend.python.inputs.InputList): A list of inputs that
        will be served to the algorithm.

      output_list (beat.backend.python.outputs.OutputList): A list of outputs
        that the algorithm will produce.

      data_sources (list): A list with all data-sources created by our execution
        loader.

      data_sinks (list): A list with all data-sinks created by our execution
        loader. These are useful for clean-up actions in case of problems.

    """

    def __init__(self, host, prefix, data, cache=None, dataformat_cache=None,
            database_cache=None, algorithm_cache=None, library_cache=None,
            custom_root_folders=None):

        super(DockerExecutor, self).__init__(prefix, data, host.ip, cache=cache,
                                             dataformat_cache=dataformat_cache,
                                             database_cache=database_cache,
                                             algorithm_cache=algorithm_cache,
                                             library_cache=library_cache,
                                             custom_root_folders=custom_root_folders)

        # Initialisations
        self.host = host


    def process(self, virtual_memory_in_megabytes=0, max_cpu_percent=0,
                timeout_in_minutes=0):
        """Executes the user algorithm code using an external program.

        The execution interface follows the backend API as described in our
        documentation.

        We use green subprocesses this implementation. Each co-process is
        linked to us via 2 uni-directional pipes which work as datain and
        dataout end-points. The parent process (i.e. the current one)
        establishes the connection to the child and then can pass/receive
        commands, data and logs.

        Usage of the data pipes (datain, dataout) is **synchronous** - you send
        a command and block for an answer. The co-process is normally
        controlled by the current process, except for data requests, which are
        user-code driven.  The nature of our problem does not require an
        *asynchronous* implementation which, in turn, would require a much more
        complex set of dependencies (on asyncio or Twisted for example).


        Parameters:

          virtual_memory_in_megabytes (:py:class:`int`, Optional): The amount
            of virtual memory (in Megabytes) available for the job. If set to
            zero, no limit will be applied.

          max_cpu_percent (:py:class:`int`, Optional): The maximum amount of
            CPU usage allowed in a system. This number must be an integer
            number between 0 and ``100*number_of_cores`` in your system. For
            instance, if your system has 2 cores, this number can go between 0
            and 200. If it is <= 0, then we don't track CPU usage.

          timeout_in_minutes (:py:class:`int`, Optional): The number of minutes
            to wait for the user process to execute. After this amount of time,
            the user process is killed with ``signal.SIGKILL``. If set to zero,
            no timeout will be applied.


        Returns:

          dict: A dictionary which is JSON formattable containing the summary
          of this block execution.

        """

        if not self.valid:
            raise RuntimeError("execution information is bogus:\n  * %s" % \
                    '\n  * '.join(self.errors))


        # Determine the docker image to use for the processing
        processing_environment = '%(name)s (%(version)s)' % self.data['environment']
        if processing_environment not in self.host:
            raise RuntimeError("Environment `%s' is not available on docker " \
                "host `%s' - available environments are %s" % (processing_environment,
                  self.host, ", ".join(self.host.processing_environments.keys())))


        # Creates the message handler
        algorithm_container = None

        def _kill():
            self.host.kill(algorithm_container)

        address = self.host.ip
        port_range = self.data.pop('port_range', None)
        if port_range:
            min_port, max_port = port_range.split(':')
            port = utils.find_free_port_in_range(int(min_port), int(max_port))
            address +=  ':{}'.format(port)

        self.message_handler = message_handler.MessageHandler(address,
                                                              kill_callback=_kill)


        #----- (If necessary) Instantiate the docker container that provide the databases

        databases_container = None
        databases_configuration_path = None
        datasets_uid =  self.data.pop('datasets_uid', None)
        network_name = self.data.pop('network_name', 'bridge')

        if len(self.databases) > 0:

            # Configuration and needed files
            databases_configuration_path = utils.temporary_directory()
            self.dump_databases_provider_configuration(databases_configuration_path)

            # Modify the paths to the databases in the dumped configuration files
            root_folder = os.path.join(databases_configuration_path, 'prefix', 'databases')

            database_paths = {}

            for db_name in self.databases.keys():
                json_path = os.path.join(root_folder, db_name + '.json')

                with open(json_path, 'r') as f:
                    db_data = simplejson.load(f)

                database_paths[db_name] = db_data['root_folder']
                db_data['root_folder'] = os.path.join('/databases', db_name)

                with open(json_path, 'w') as f:
                    simplejson.dump(db_data, f, indent=4)

            # Determine the docker image to use for the databases
            try:
                databases_environment = self.host.db2docker(database_paths.keys())
            except:
                raise RuntimeError("No environment found for the databases `%s' " \
                    "- available environments are %s" % (
                      ", ".join(database_paths.keys()),
                      ", ".join(self.host.db_environments.keys())))

            # Creation of the container
            # Note: we only support one databases image loaded at the same time
            database_port = utils.find_free_port()

            cmd = [
                'databases_provider',
                '0.0.0.0:%d' % database_port,
                '/beat/prefix',
                '/beat/cache'
            ]

            if logger.getEffectiveLevel() <= logging.DEBUG:
                cmd.insert(1, '--debug')

            databases_container_name = "beat_db_%s" % utils.id_generator()
            databases_container = self.host.create_container(databases_environment, cmd)
            databases_container.uid = datasets_uid
            databases_container.network_name = network_name
            databases_container.set_name(databases_container_name)

            # Specify the volumes to mount inside the container
            databases_container.add_volume(databases_configuration_path, '/beat/prefix')
            databases_container.add_volume(self.cache, '/beat/cache')

            for db_name, db_path in database_paths.items():
                databases_container.add_volume(db_path, os.path.join('/databases', db_name))

            # Start the container
            while True:
                try:
                    databases_container.add_port(database_port, database_port, host_address=self.host.ip)

                    self.host.start(databases_container)

                    break
                except Exception as e:
                    if str(e).find('port is already allocated') < 0:
                        break

                    databases_container.reset_ports()
                    database_port = utils.find_free_port()

                    cmd = [x if not x.startswith('0.0.0.0:') else '0.0.0.0:%d' % database_port for x in cmd]
                    databases_container.command = cmd

            database_ip = self.host.get_ipaddress(databases_container)


        #----- Instantiate the algorithm container

        # Configuration and needed files
        configuration_path = utils.temporary_directory()
        self.dump_runner_configuration(configuration_path)

        # Command to execute
        cmd = [
            'execute',
            '--cache=/beat/cache',
            self.message_handler.address,
            '/beat/prefix'
        ]

        if len(self.databases) > 0:
            cmd.append('tcp://' + database_ip + ':%d' % database_port)

        if logger.getEffectiveLevel() <= logging.DEBUG:
            cmd.insert(1, '--debug')

        # Creation of the container
        algorithm_container = self.host.create_container(processing_environment, cmd)
        algorithm_container.uid = datasets_uid
        algorithm_container.network_name = network_name

        # Volumes
        algorithm_container.add_volume(configuration_path, '/beat/prefix')
        algorithm_container.add_volume(self.cache, '/beat/cache', read_only=False)

        # Start the container
        self.host.start(algorithm_container,
                        virtual_memory_in_megabytes=virtual_memory_in_megabytes,
                        max_cpu_percent=max_cpu_percent
        )


        # Process the messages until the container is done
        self.message_handler.start()

        timed_out = False

        try:
            timeout = (60 * timeout_in_minutes) if timeout_in_minutes else None
            status = self.host.wait(algorithm_container, timeout)

        except requests.exceptions.ReadTimeout:
            logger.warn("user process has timed out after %d minutes", timeout_in_minutes)
            timed_out = True

            self.host.kill(algorithm_container)
            status = self.host.wait(algorithm_container)

        except KeyboardInterrupt: # Developer pushed CTRL-C
            logger.info("stopping user process on CTRL-C console request")
            self.host.kill(algorithm_container)
            status = self.host.wait(algorithm_container)

        finally:
            if databases_container is not None:
                self.host.kill(databases_container)
                self.host.wait(databases_container)

            self.message_handler.stop.set()
            self.message_handler.join()


        # Collects final information and returns to caller
        container_log = self.host.logs(algorithm_container)

        if status != 0:
            stdout = ''
            stderr = container_log
        else:
            stdout = container_log
            stderr = ''

        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug("Log of the container: " + container_log)

        retval = dict(
            status = status,
            stdout = stdout,
            stderr = stderr,
            timed_out = timed_out,
            statistics = self.host.statistics(algorithm_container),
            system_error = self.message_handler.system_error,
            user_error = self.message_handler.user_error,
        )

        retval['statistics']['data'] = self.message_handler.statistics
        stats.update(retval['statistics']['data'], self.io_statistics)

        self.host.rm(algorithm_container)

        if databases_container is not None:
            db_container_log = self.host.logs(databases_container)

            if logger.getEffectiveLevel() <= logging.DEBUG:
                logger.debug("Log of the database container: " + db_container_log)

            if status != 0:
                retval['stderr'] += '\n' + db_container_log
            else:
                retval['stdout'] += '\n' + db_container_log

            self.host.rm(databases_container)

        self.message_handler.destroy()
        self.message_handler = None

        if not self.debug:
            if databases_configuration_path:
                shutil.rmtree(databases_configuration_path)
            shutil.rmtree(configuration_path)

        return retval
